//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* mystd.cpp: feeble attempts to cure the many deficiencies of C++        *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h> // otherwise buggy compiler error from gsl
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <glob.h>
#include <iostream>

#include "mystd.h"


//! declare auxiliary functions.

namespace mystd_aux {
    int tim2dbl(const string inp, double *val);
};


//**************************************************************************//
//*  constants                                                             *//
//**************************************************************************//

const char mystd::BEL = 0x7;
const double mystd::PI = 3.1415926535897931;


//**************************************************************************//
//*  convert to string                                                     *//
//**************************************************************************//

string strg(const string& val)
{
    return val;
}

string strg(const bool& val) {
    if (val)
        return "y";
    else
        return "n";
}

string strg(const double& val)
{
    char s[40];
    sprintf(s, "%g", val);
    return s;
}

string strg(const int& val)
{
    char s[40];
    sprintf(s, "%d", val);
    return s;
}

string strg(const uint& val)
{
    char s[40];
    sprintf(s, "%u", val);
    return s;
}

string strg(const unsigned long& val)
{
    char s[40];
    sprintf(s, "%lu", val);
    return s;
}

string strg(const char& val)
{
    return string(1, val);
}


//**************************************************************************//
//*  string operations                                                     *//
//**************************************************************************//

string mystd::strip(
    string const& str, ///< the original string
    char const* sepSet ///< C string with characters to be dropped
)
{
    string::size_type const first = str.find_first_not_of(sepSet);
    return ( first==std::string::npos )
        ? std::string()
        : str.substr(first, str.find_last_not_of(sepSet)-first+1);
}

//**************************************************************************//
//*  yaml                                                                  *//
//**************************************************************************//

//! refine string for yaml output.

string yaml( const string s )
// string -> yaml-output-string
{
    string out = s;
    size_t i = 0;
    // escape "
    while( i < out.size() ){
        if( out[i]=='"' && (i==0 || out[i-1]!='\\') )
            out.insert( i, "\\" );
        ++i;
    }
    // if there are special characters, put ".." around
    if( s.find( '"' )!=string::npos ||
        s.find( ':' )!=string::npos ||
        s.find( '*' )!=string::npos ||
        s.find( '&' )!=string::npos ||
        s.find( '>' )!=string::npos ||
        s.find( '<' )!=string::npos
        )
        out = '"' + out + '"';
    return out;
}


//**************************************************************************//
//*  convert from string                                                   *//
//**************************************************************************//

bool mystd::str2vec( string inp, vector<double> *V, uint nmax, bool force )
{
    using namespace mystd_aux;
    using namespace std;

    // cout << "DEBUG mystd [" << inp << "]\n";
    V->clear();

    while(inp.size() && isspace(inp[0])) inp.erase(inp.begin());
    // while(inp.size() && strchr(" \t.\n",inp[inp.size()-1])) 
    //    inp.erase(inp.end()-1);

    // cout << "DEBUG  core [" << inp << "]\n";
    if (inp==string("")) {
        // cout << "DEBUG  empty\n";
        return true;
    }
    // cout << "DEBUG  nonempty\n";

    string::size_type ji=0, jf;
    double val;
    char dummy;
    while (ji<inp.size()) {
        if ((jf=inp.find_first_of(", \t\n", ji)) == string::npos) 
            jf = inp.size();
        // cout << " look until " << jf << "\n";
        if (sscanf(inp.substr(ji,jf-ji).c_str(), "%lg%c", &val, &dummy)
            ==1) {
            V->push_back(val);
        } else if ( force ) {
            cout << "! str2vec: invalid value ["
                 << inp.substr(ji,jf-ji) << "] at pos " << ji
                 << " in [" << inp << "]\n";
            return false;
        }
        if (nmax && V->size()>=nmax) break; // do not decode the tail
        // cout << " found "<<ji<<":"<<jf<<" value "<<val<<"\n";
        ji = inp.find_first_not_of(", \t\n", jf);
        // cout << " continue at " << ji << "\n";
    }

    return true;
}

bool mystd::any2dbl( const string inp, double *val )
{
    if(!inp.size()) {
        cout << "expecting double, found empty input\n";
        return false;
    }
    if( inp=="-inf" || inp=="-oo" ){
        *val = -INFINITY;
        return false;
    }
    if( inp=="inf" || inp=="oo" || inp=="+inf" || inp=="+oo" ){
        *val = INFINITY;
        return false;
    }
    if(!strchr("+-0123456789.", inp[0])) {
        cout << "expecting double, found bad char[0] in [" 
             << inp << "]\n";
        return false;
    }
    double mant; int expt;
    if (sscanf(inp.c_str(), "%lgm%d", &mant, &expt)==2) {
        *val = mant * pow(10.e0, (double)(-expt));
        return true;
    }
    if (sscanf(inp.c_str(), "%lg", val)==1)
        return true;
    cout << "expecting double, sscanf failed\n";
    return false;
}

bool mystd::any2int(const string inp, int *val)
{
    if(!inp.size()) {
        cout << "expecting integer, found empty input\n";
        return false;
    }
    if(!strchr("+-0123456789", inp[0])) {
        cout << "expecting integer, found bad char[0] in [" 
             << inp << "]\n";
        return false;
    }
    if (sscanf(inp.c_str(), "%d", val)==1) return true;
    cout << "expecting integer, sscanf failed\n";
    return false;
}

bool mystd::any2uint(const string inp, uint *val)
{
    if(!inp.size()) {
        cout << "expecting integer, found empty input\n";
        return false;
    }
    if(!strchr("0123456789", inp[0])) {
        cout << "expecting integer, found bad char[0] in [" 
             << inp << "]\n";
        return false;
    }
    if (sscanf(inp.c_str(), "%u", val)==1) return true;
    cout << "expecting integer, sscanf failed\n";
    return false;
}

void mystd::string_extract_word( const string in, string *out1, string *out2 )
{
    // one word (*out1) is extracted; leading " \t" are retained.
    // in the tail (*out2) leading " \t" are deleted.
    // "\n" starts always a new word.
    // leading "`" makes *out1 go to next "`" or to end of line.

    string::size_type j0=0, j1, j2, jf=in.size();

    // handle special case "`"
    if (in[0]=='`') { 
        j1 = in.find_first_of('`', 1);
        if (j1==string::npos) {
            *out1 = in.substr(1);
            *out2 = "";
        } else {
            *out1 = in.substr(1,j1-1);
            j2 = in.find_first_not_of(" \t", j1+1);
            if (j2!=string::npos) 
                *out2 = in.substr(j2);
            else
                *out2 = "";
        }
        return;
    }

    while (j0<jf && strchr(" \t",in[j0])) j0++;
    j1=j0+1;
    while (j1<jf && !strchr(" \t\n",in[j1])) j1++;
    j2=j1;
    while (j2<jf && strchr(" \t",in[j2])) j2++;
    *out1 = in.substr(0,j1);
    if (j2<jf) 
        *out2 = in.substr(j2);
    else
        *out2 = string("");
    //cout << "from ["<<in<<"] word ["<<*out1<<"] and tail ["<<*out2<<"]\n";
}

void mystd::string_extract_line(const string in, string *out1, string *out2)
{
    string line;
    string::size_type i = in.find('\n');
    if (i==string::npos) {
        *out1 = in;
        *out2 = string("");
    } else {
        *out1 = in.substr(0,i);
        *out2 = in.substr(i+1);
    }
    //cout << "from ["<<in<<"] line ["<<*out1<<"] and tail ["<<*out2<<"]\n";
}

//**************************************************************************//
//*  improved libc calls                                                   *//
//**************************************************************************//

void mystd::system( string cmd )
{
    int ret = ::system( cmd.c_str() );
    if( ret )
        throw "Command '" + cmd + "' returned error code " + strg(ret);
}

//**************************************************************************//
//*  time                                                                  *//
//**************************************************************************//

//! Seconds since the epoch. Return them as double (WHY??)
//  only used for calls to myasctime => OBSOLETE

double mystd::justnow(void)
{
    struct timeval exactlynow;
    gettimeofday(&exactlynow, NULL);
    return (double) exactlynow.tv_sec;
}

//! Return time as string in ctime/asctime standard format.
//  OBSOLETE: better use strftime

string mystd::myasctime(double time, int offs)
{
    long timin = (long) time;
    char buf[40];
    sprintf(buf, "%.24s", asctime(localtime(&timin)));
    return string(buf+offs);
}


//**************************************************************************//
//*  rounding                                                              *//
//**************************************************************************//

//! Round given value to given number of digits.

double mystd::round_decimal(double val, double digits)
{
    if (val==0 || digits<=0) return 0;
    double v = fabs(val);
    int n = (int) floor ( digits - log10(v));
    // cout << " n=" << n;
    double p = pow(10., n);
    double r = round(v*p) / p;
    return (val<0 ? -1 : +1) * r;
}


//**************************************************************************//
//*  vector operations                                                     *//
//**************************************************************************//

//! insert val into vector *V, presuming the latter is sorted

void mystd::merge(vector<double> *V, double val)
{
    for (uint i=0; i<V->size(); ++i) {
        if (val>(*V)[i]) continue;
        if (val==(*V)[i]) return;
        V->insert(V->begin()+i, val);
        return;
    }
    V->push_back(val);
}

/*
bool mystd::get_const(vector<double> *V, double *val)
    // check whether *V has a constant value; write this value to *val
{
    double v;
    if (V->size()==0) return false; // empty vector
    v = (*V)[0];
    for (uint i=1; i<V->size(); ++i)
        if ((*V)[i]!=v) return false; // vector not constant
    if (val) *val = v;
    return true;
}
*/

void mystd::unique(vector<double> *V, double tolabs, double tolrel)
{
    // tolrel z.Zt. nicht benutzt
    vector<double> aux;
    double diff;
    if (!V->size()) return;
    aux.push_back((*V)[0]);
    for (uint ii=1; ii<V->size(); ++ii)
        if((diff=fabs((*V)[ii]-aux.back()))>tolabs)
            aux.push_back((*V)[ii]);
    *V = aux;
}

//! detect sorting (strictly de- or increasing: -1 or1; else: 0).

int mystd::sorted( const vector<double>& V )
{
    uint n = V.size();
    if( n < 2 )
        throw string( "vector too short to determine sorting" );
    int ret;
    if      ( V[n-1] > V[0] )
        ret = 1;
    else if ( V[n-1] < V[0] )
        ret = -1;
    else
        return 0;
    for( uint i=2; i<n; ++i )
        if ( ret*(V[i]-V[i-1]) <= 0 )
            return 0;
    return ret;
}

//! detect equidistant grid, set *step.

bool mystd::is_equidist( double *step, const vector<double>& V )
{
    uint n = V.size();
    if( n < 2 )
        throw string( "vector too short to determine equidistance" );
    *step = ( V[n-1] - V[0] ) / (n-1);
    for( uint i=1; i<n-1; ++i )
        if( fabs( V[i] - (V[0]+i*(*step)) ) > 1e-12*fabs(*step)+1e-200 )
            return false;
    return true;
}

//! sum over vector elements.

double mystd::sum( const vector<double>& V )
{
    double ret = 0;
    for( uint i=0; i<V.size(); ++i )
        ret += V[i];
    return ret;
}

//**************************************************************************//
//*  file names, globbing, file read                                       *//
//**************************************************************************//

void mystd::fname_divide( const string fname,
                          string *fdir, string *fshort, string *fext )
// divide "dir/short.ext" into "dir", "short", "ext"
// if some of the three return values are not needed, the user may supply 0
// instead of a pointer to a string.
{
    uint idir = fname.rfind("/") + 1;
    if( fdir ) *fdir = fname.substr( 0, idir );

    uint iext = fname.rfind(string(".")) + 1;
    if( iext<=idir ){
        if( fext ) *fext = "";
        if( fshort ) *fshort = fname.substr( idir );
    } else {
        if( fext ) *fext = fname.substr( iext );
        if( fshort ) *fshort = fname.substr( idir, iext-idir-1 );
    }
}

int mystd::glob_file_list( const string text, 
                           const string pre, const string ext,
                           vector<string> *flong, vector<string> *fshort )
// input:
//     text:   bash-like file list
//     pre:    prepend this
//     ext:    append this (except if text already contains extension)
// output:
//     flong:  unix file names
//     fshort: internal (rda) file names  

// the implementation with globpattern is OBSOLET: better use wordexp(3)

{
    string globpattern;
    if (pre!=string("")) 
        globpattern = pre + string("/") + text;
    else
        globpattern = text;
    if( ext!=string("") and
        not ( (int)(text.rfind(".")) > 0 and (int)(text.rfind(".")) > (int)(text.rfind("/")) ) ) 
        globpattern += string(".") + ext;

    glob_t globbuf;
    glob(globpattern.c_str(), GLOB_TILDE, NULL, &globbuf);
    // printf( "DEBUG: text[%s] -> globpattern[%s] -> count[%i] because pos(.)=%i, pos(/)=%i\n",
    //        text.c_str(), globpattern.c_str(), globbuf.gl_pathc, text.rfind("."), text.rfind("/") );

    for (unsigned int i=0; i<globbuf.gl_pathc; i++) {
        string fnam = globbuf.gl_pathv[i];
        uint ipre = fnam.rfind("/") + 1;
        uint isub = fnam.rfind(string(".")+ext) - ipre;
        flong->push_back(fnam);
        if( fshort ) fshort->push_back(fnam.substr(ipre, isub));
    }

    globfree(&globbuf);
    return flong->size();
}

int mystd::glob_one_file(const string text, const string pre, const string ext, 
                         string *fname)
{
    vector<string> flong, fshort;
    int ret = glob_file_list(text, pre, ext, &flong, &fshort);
    if (ret>=1) *fname = flong[0];
    return ret;
}

FILE* mystd::glob_fopen(const string text, const string pre, const string ext, 
                        const char* mod, string *fname)
{
    string flong;
    if (glob_one_file(text, pre, ext, &flong)!=1) return 0;
    if (fname) *fname=flong;
    return fopen(flong.c_str(), mod);
}

int mystd::freadln( FILE *fd, string *s )
{
    char c = fgetc(fd);
    if( c==EOF )
        return -1;
    if( c=='\n' ){
        *s = "";
        return 0;
    }
    *s = c;
    while ( 1 ) {
        c = fgetc(fd);
        if ( c=='\n' || c==EOF )
            return s->size();
        *s += c;
    }
}

//**************************************************************************//
//*  GSL integration                                                       *//
//**************************************************************************//

#include <gsl/gsl_integration.h>

double mystd::Integrate( double (*func_) (double, void*), void *data_, 
                         const double low, const double hig,
                         const uint mode, const double epsabs,
                         const double epsrel )
{
    gsl_function F;
    F.function = func_;
    F.params = data_;
#define nwork 20000
    gsl_integration_workspace *work = gsl_integration_workspace_alloc( nwork );
    double val, err;
    size_t neval;

    if      ( isinf(low)==-1 && isinf(hig)==+1 )
        gsl_integration_qagi( &F, 
                             epsabs, epsrel, nwork,
                             work, &val, &err );
    else if ( isinf(low)==-1 )
        gsl_integration_qagil( &F, hig,
                             epsabs, epsrel, nwork,
                             work, &val, &err );
    else if ( isinf(hig)==+1 )
        gsl_integration_qagiu( &F, low,
                             epsabs, epsrel, nwork,
                             work, &val, &err );
    else if ( mode==0 )
        // QNG non-adaptive Gauss-Kronrod integration
        gsl_integration_qng( &F, low, hig,
                             epsabs, epsrel,
                             &val, &err, &neval );
    else if ( mode>0 && mode<7 ){
        int rule;
        switch( mode ){
        case 1: rule = GSL_INTEG_GAUSS61; break;
        case 2: rule = GSL_INTEG_GAUSS51; break;
        case 3: rule = GSL_INTEG_GAUSS41; break;
        case 4: rule = GSL_INTEG_GAUSS31; break;
        case 5: rule = GSL_INTEG_GAUSS21; break;
        case 6: rule = GSL_INTEG_GAUSS15; break;
        }
        // QAG adaptive integration
        gsl_integration_qag( &F, low, hig,
                             epsabs, epsrel, nwork,
                             rule, work, &val, &err );
    }
    else if ( mode==7 )
        // QAGS adaptive integration with singularities
        gsl_integration_qags( &F, low, hig,
                              epsabs, epsrel, nwork,
                              work, &val, &err );
    else
        throw string( "invalid mode in mystd::Integrate" );

    gsl_integration_workspace_free( work );
    return val;
}
