//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* scan.cpp: each online data file contains at least one scan             *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>

#include "mystd.h"
#include "scan.h"

//**************************************************************************//
//*  Construct or modify                                                   *//
//**************************************************************************//

//! Append given data point.

void CScan::push_back( const double x_in, const double y_in )
{
    x.push_back(x_in);
    y.push_back(y_in);
}

//! Remove data points according to given indices.

void CScan::erase( const uint begin, const uint end )
{
    x.erase( x.begin()+begin, x.begin()+end );
    y.erase( y.begin()+begin, y.begin()+end );
}

//! Reset to initial state.

void CScan::Clear(void)
{
    x.clear();
    y.clear();
    z.clear();
}

//**************************************************************************//
//*  Extract data                                                          *//
//**************************************************************************//

//! Return number of data points.

size_t CScan::size() const
{
    if ( x.size()!=y.size() )
        throw "SEVERE BUG detected by scan::size: #x [" + 
                  strg(x.size()) + "] <> #y [" + strg(y.size()) + "]\n";
    return y.size();
}

//! For each xx[i], set yy[i] by interpolating in this.x -> this.y.

void CScan::intpol( const vector<double>& xx, vector<double>& yy ) const
{
    size_t n = size();
    size_t nn = xx.size();
    yy.resize( nn );

    size_t ii = 0;
    while( xx[ii]<x[0] ){
        if( ii>=nn )
            return;
        yy[ii] = 0;
        ++ii;
    }
    size_t i = 0;
    while ( i<n-1 && xx[ii]<x[n-1] ) {
        while ( xx[ii]>x[i+1] && i<n-1 )
            ++i;
        if( i>=n-1 )
            break;
        if( ii>=nn )
            return;
        yy[ii] = y[i] + (xx[ii]-x[i])/(x[i+1]-x[i])*(y[i+1]-y[i]);
        ++ii;
    }
    for( ; ii<nn; ++ii )
        yy[ii] = 0;
}

//**************************************************************************//
//*  Operations outside CScan                                              *//
//**************************************************************************//

//! Compare z vectors of two scans, for use within sort routine.

bool CompareZ( const CScan& S1, const CScan& S2 )
{
    if( S1.z.size()!=S2.z.size() )
        return false;
    size_t nz = S1.z.size();
    for( size_t iz=0; iz<nz; ++iz ){
        if    ( S1.z[iz]>S2.z[iz] )
            return false;
        if    ( S1.z[iz]<S2.z[iz] )
            return true;
    }
    return false;
}
